unit mtph2;

{Algorithm Description:
 Step 1: Uppercase and remove all non-alpha characters
 Step 2: Scan for phase one rules
 Step 3: Remove all double characters
 Step 4: Scan for phase two rules
 Step 5: Return value
}

{ Version    Comments
   1.5b      Beta for upcoming version 2 - complete rewrite

}

interface

	function metaphone(a : PChar; lg : integer) : PChar; stdcall;

implementation

uses classes, sysutils;

type
	TRuleType =
   (mrBeginningOfWord,    //if FromStr is at begining of word
    mrAfter,              //If FromStr is just after arg
    mrBefore,             //If FromStr is just before arg
    mrNotAfterVowel,      //If FromStr is not preceded by vowel
    mrNotBeforeVowel,     //If FromStr is not followed by vowel
    mrBetween,            //If FromStr is between two chars specified in Args
    mrNotEndAfter,        //Not at end of word after string in Args
    mrNotEndBefore,       //Not at end of word just before string in args
    mrBeforeNoVowel,      //Before "args", but no vowel thereafter
    mrAfterVowelNotBeforeVowel, //after, but not before vowel
    mrAtEndBefore,        //At end of word just bofore "Arg"
    mrNone);              //Rule always applies

  TRule = record
    FromStr: string;
    ToStr: string;
    RuleType: TRuleType;
    Args: string;
  end;

const
  RuleCount = 67;

  //note - always make the default rule the LAST
  Rules: array [1..RuleCount] of TRule = (
    //beginning of word rules
   (FromStr: ('AE');ToStr: ('E');RuleType: mrBeginningOfWord; Args: ('')),
   (FromStr: ('GN');ToStr: ('N');RuleType: mrBeginningOfWord; Args: ('')),
   (FromStr: ('KN');ToStr: ('N');RuleType: mrBeginningOfWord; Args: ('')),
   (FromStr: ('PN');ToStr: ('N');RuleType: mrBeginningOfWord; Args: ('')),
   (FromStr: ('WR');ToStr: ('R');RuleType: mrBeginningOfWord; Args: ('')),
   (FromStr: ('PS');ToStr: ('S');RuleType: mrBeginningOfWord; Args: ('')),
   (FromStr: ('WH');ToStr: ('W');RuleType: mrBeginningOfWord; Args: ('')), //8 rules

   //B
   (FromStr: ('B');ToStr: ('B');RuleType: mrNotEndAfter; Args: ('M')),
   (FromStr: ('B');ToStr: ('B');RuleType: mrNone; Args: ('')),

   //C
   (FromStr: ('C');ToStr: ('X');RuleType: mrBetween; Args: ('CA')),
   (FromStr: ('C');ToStr: ('X');RuleType: mrBefore; Args: ('H')),

   (FromStr: ('C');ToStr: ('S');RuleType: mrBefore; Args: ('I')),
   (FromStr: ('C');ToStr: ('S');RuleType: mrBefore; Args: ('E')),
   (FromStr: ('C');ToStr: ('S');RuleType: mrBefore; Args: ('Y')),

   (FromStr: ('C');ToStr: ('');RuleType: mrBetween; Args: ('SE')),
   (FromStr: ('C');ToStr: ('');RuleType: mrBetween; Args: ('SI')), //10

   (FromStr: ('C');ToStr: ('K');RuleType: mrNone; Args: ('')),

    //D
   (FromStr: ('D');ToStr: ('J');RuleType: mrBefore; Args: ('GE')),
   (FromStr: ('D');ToStr: ('J');RuleType: mrBefore; Args: ('GY')),
   (FromStr: ('D');ToStr: ('J');RuleType: mrBefore; Args: ('GI')),

   (FromStr: ('D');ToStr: ('T');RuleType: mrNone; Args: ('')),

    //F
   (FromStr: ('F');ToStr: ('F');RuleType: mrNone; Args: ('')),

    //GG - changed to "1" in phase 1
   (FromStr: ('1');ToStr: ('K');RuleType: mrNone; Args: ('')),

   //G
   (FromStr: ('G');ToStr: ('G');RuleType: mrBefore; Args: ('G')),
   (FromStr: ('G');ToStr: ('');RuleType: mrAfter; Args: ('G')),
   (FromStr: ('G');ToStr: ('');RuleType: mrBeforeNoVowel; Args: ('H')),
   (FromStr: ('G');ToStr: ('');RuleType: mrAtEndBefore; Args: ('N')),
   (FromStr: ('G');ToStr: ('');RuleType: mrAtEndBefore; Args: ('NED')),
   (FromStr: ('G');ToStr: ('');RuleType: mrBetween; Args: ('DE')), //paired with D rule

   (FromStr: ('G');ToStr: ('J');RuleType: mrBefore; Args: ('I')),
   (FromStr: ('G');ToStr: ('J');RuleType: mrBefore; Args: ('E')),
   (FromStr: ('G');ToStr: ('J');RuleType: mrBefore; Args: ('Y')),

   (FromStr: ('G');ToStr: ('K');RuleType: mrNone; Args: ('')), //16

   //H
   (FromStr: ('H');ToStr: ('');RuleType: mrAfterVowelNotBeforeVowel; Args: ('')),
   (FromStr: ('H');ToStr: ('');RuleType: mrAfter; Args: ('C')),
   (FromStr: ('H');ToStr: ('');RuleType: mrAfter; Args: ('S')),
   (FromStr: ('H');ToStr: ('');RuleType: mrAfter; Args: ('P')),
   (FromStr: ('H');ToStr: ('');RuleType: mrAfter; Args: ('T')),
   (FromStr: ('H');ToStr: ('');RuleType: mrAfter; Args: ('G')),

   (FromStr: ('H');ToStr: ('');RuleType: mrNone; Args: ('')),

   //J
   (FromStr: ('J');ToStr: ('J');RuleType: mrNone; Args: ('')),

    //K
   (FromStr: ('K');ToStr: ('');RuleType: mrAfter; Args: ('C')),

   (FromStr: ('K');ToStr: ('K');RuleType: mrNone; Args: ('')), //10

    //L
   (FromStr: ('L');ToStr: ('L');RuleType: mrNone; Args: ('')),

   //M
   (FromStr: ('M');ToStr: ('M');RuleType: mrNone; Args: ('')),

   //N
   (FromStr: ('N');ToStr: ('N');RuleType: mrNone; Args: ('')),

   //P
   (FromStr: ('P');ToStr: ('F');RuleType: mrBefore; Args: ('H')),

   (FromStr: ('P');ToStr: ('P');RuleType: mrNone; Args: ('')),

   //Q
   (FromStr: ('Q');ToStr: ('K');RuleType: mrNone; Args: ('')),

    //R
   (FromStr: ('R');ToStr: ('R');RuleType: mrNone; Args: ('')),

   //S
   (FromStr: ('S');ToStr: ('X');RuleType: mrBefore; Args: ('H')),
   (FromStr: ('S');ToStr: ('X');RuleType: mrBetween; Args: ('SO')),
   (FromStr: ('S');ToStr: ('X');RuleType: mrBetween; Args: ('SA')),

   (FromStr: ('S');ToStr: ('S');RuleType: mrNone; Args: ('')), //11

   //T
   (FromStr: ('T');ToStr: ('X');RuleType: mrBefore; Args: ('IA')),
   (FromStr: ('T');ToStr: ('X');RuleType: mrBefore; Args: ('IO')),

   (FromStr: ('T');ToStr: ('');RuleType: mrBefore; Args: ('CH')),
   (FromStr: ('T');ToStr: ('0');RuleType: mrBefore; Args: ('H')),

   (FromStr: ('T');ToStr: ('T');RuleType: mrNone; Args: ('')),

    //V
   (FromStr: ('V');ToStr: ('F');RuleType: mrNone; Args: ('')),

   //W
   (FromStr: ('W');ToStr: ('W');RuleType: mrNotBeforeVowel; Args: ('')),

   (FromStr: ('W');ToStr: ('');RuleType: mrNone; Args: ('')),

   //X
   (FromStr: ('X');ToStr: ('S');RuleType: mrBeginningOfWord; Args: ('')),
   (FromStr: ('X');ToStr: ('KS');RuleType: mrNone; Args: ('')),

   //Y
   (FromStr: ('Y');ToStr: ('Y');RuleType: mrNotBeforeVowel; Args: ('')),

   (FromStr: ('Y');ToStr: ('');RuleType: mrNone; Args: ('')),

   //Z
   (FromStr: ('Z');ToStr: ('S');RuleType: mrNone; Args: (''))
   ); //12

  Vowels='AEIOU';
  AllowChar='ABCDEFGHIJKLMNOPQRSTUVWXYZ';
var
  OutStr: TStringList;

procedure ApplyRules(ScanStr: string; RuleBase: array of TRule);

  function RuleApplicable(Rule: TRule; CharIndex: integer): boolean;
  var
    RemChar, PrevChar, ArgLength, InpLength: integer;
  begin
    InpLength:=length(ScanStr);
    RemChar:=InpLength-CharIndex;
    PrevChar:=CharIndex-1;
    ArgLength:=length(Rule.Args);

    result:=false;

    if ((RemChar+1)>=length(Rule.FromStr)) and
      (copy(ScanStr,CharIndex,length(Rule.FromStr))=Rule.FromStr) then

      case Rule.RuleType of
        mrBeginningOfWord:    //if FromStr is at beggining of word
          if CharIndex=1 then
            result:=true;

        mrAfter:              //If FromStr is just after arg
          if (PrevChar>=ArgLength) and
            (copy(ScanStr,CharIndex-ArgLength-1,ArgLength)=Rule.Args) then
              result:=true;

        mrBefore:             //If FromStr is just before arg
          if (RemChar>=ArgLength) and
            (copy(ScanStr,CharIndex+1,ArgLength)=Rule.Args) then
              result:=true;

        mrNotAfterVowel:      //If FromStr is not preceded by vowel
          if (RemChar>=1) and (pos(copy(ScanStr,CharIndex+1,1),Vowels)>0) then
            result:=true;

        mrNotBeforeVowel:     //If FromStr is not followed by vowel
          if (PrevChar>=1) and (pos(copy(ScanStr,CharIndex+1,1),Vowels)>0) then
            result:=true;

        mrBetween:            //If FromStr is between two chars specified in Args
          if (PrevChar>=1) and (RemChar>=1) and (length(rule.args)=2) and
            (copy(ScanStr,CharIndex-1,1)=copy(rule.args,1,1)) and
            (copy(ScanStr,CharIndex+1,1)=copy(rule.args,2,1)) then
              result:=true;

        mrNotEndAfter:        //Not at end of word after string in Args
          if (CharIndex<InpLength) and (PrevChar>=length(rule.args)) and
            (copy(ScanStr,CharIndex-ArgLength-1,ArgLength)=Rule.Args) then
              result:=true;

        mrNotEndBefore:       //Not at end of word just before string in args
          if (ArgLength>RemChar) and
            (copy(ScanStr,CharIndex+1,ArgLength)=Rule.Args) then
              result:=true;

        mrBeforeNoVowel:      //Before "args", but no vowel thereafter
          if (ArgLength+1<=RemChar) and
            (copy(ScanStr,CharIndex+1,ArgLength)=Rule.Args) and
            (pos(vowels,copy(ScanStr,CharIndex+1+ArgLength,1))>0) then
              result:=true;

        mrAfterVowelNotBeforeVowel: //after, but not before vowel
          if (PrevChar>0) and (RemChar>0) and
            (pos(Vowels,copy(ScanStr,CharIndex-1,1))>0) and
            (pos(Vowels,copy(ScanStr,CharIndex+1,1))=0) then
              result:=true;

        mrAtEndBefore:        //At end of word just before "Arg"
          if (ArgLength=RemChar) and
            (copy(ScanStr,CharIndex+1,ArgLength)=Rule.Args) then
              result:=true;

        mrNone:               //Rule always applies
          Result:=true;
      end;  //case

  end; //function RuleApplicable

var
  t, i: integer;
  Res, SkipRule: string;
  SkipFlag: boolean;
begin
  t:=low(RuleBase);
  while t<=high(RuleBase) do begin
    SkipFlag:=false;
    for i:=1 to length(ScanStr) do
      if RuleApplicable(RuleBase[t],i) then begin
        OutStr.addobject(RuleBase[t].ToStr,pointer(i));
        SkipFlag:=true;
        SkipRule:=RuleBase[t].FromStr;
      end;
    if SkipFlag then
      while RuleBase[t].FromStr=SkipRule do
        inc(t)
    else
      inc(t); //normal increment
  end;
end;

function findrel(xx: integer): string;
var
  l: integer;
begin
  result:='';
  for l:=0 to outstr.count-1 do
    if integer(outstr.objects[l])=xx then
      result:=result+format('%3.3s %d ',[OutStr[l],l]);
end;

function metaphone(a : PChar; lg : integer) : PChar; stdcall;
var
  InStr, TempStr: string;
  x, y, SmallestIndex, SmallestValue: integer; //for selection sort
  Reslt: string;
  FirstFlag: boolean;
begin
  OutStr:=TStringList.create;
  Reslt:='';

  try
    //upper case
    TempStr:=uppercase(a);
    InStr:='';

    for x:=1 to length(TempStr) do
      if pos(copy(TempStr,x,1),AllowChar)>0 then
        InStr:=InStr+copy(TempStr,x,1);

    //remove doubles EXCEPT FOR G (ugly exception)
    if length(InStr)>0 then begin
      TempStr:=copy(instr,1,1);
      for x:=2 to length(InStr) do
        if (copy(instr,x,1)='G') then
          TempStr:=TempStr+copy(instr,x,1)
        else if (copy(instr,x,1)<>copy(instr,x-1,1)) then
          TempStr:=TempStr+copy(instr,x,1);
      InStr:=TempStr;
    end;

      //scan input and create result for each rule in output array
    ApplyRules(InStr,Rules);

    //get result - order output stringlist, then translate to string
    //do selection sort - or something like that, anyway :-)
    for x:=0 to OutStr.count-1 do begin
      SmallestIndex:=x;
      SmallestValue:=integer(OutStr.objects[x]);
      for y:=x to OutStr.count-1 do
        if integer(OutStr.objects[y])<SmallestValue then begin
          SmallestIndex:=y;
          SmallestValue:=integer(OutStr.objects[y]);
        end;
      if SmallestIndex>x then //do swap with smallest
        OutStr.exchange(x,SmallestIndex);
    end;

//    for x:=1 to length(instr) do
//      writeln(format('Org: %s Mtph: %s',[Instr[x],findrel(x)]));

    FirstFlag:=false;
    for x:=0 to OutStr.count-1 do begin
      Reslt:=Reslt+OutStr[x];
      if integer(OutStr.objects[x])=1 then
        FirstFlag:=true;
    end;

    //the following is a fix for words beginning with vowels
    //if there isn't a translation of the first character,
    //it adds whatever the first character is
    if (not firstflag) and (length(Instr)>0) then
      Reslt:=copy(Instr,1,1)+Reslt;

    result:=pchar(reslt);
//    StrLCopy(Res, pchar(Reslt), lg);
  finally
    OutStr.free;
  end;
end;

end.
